var ORM = require('./ORM');
var QueryObject = require('./QueryObject');
var Query = require('./Query');
var Relationship = require('./Relationship');

function ModelFactory() {
  this.ORM = null;
  this.Model = null;
  this.tableName = '';
  this.attributes = {};
  this.relationships = {};
  this.pre = {};
  this.post = {};
  this.static = {};
  this.method = {};
}

//
// Model Operations
//

ModelFactory.prototype.extend = function(ORM, tableName, data) {
  this.ORM = ORM;
  this.tableName = tableName;
  this.Model = this.createModel(data.method);
  this.addPre(data.pre);
  this.addPost(data.post);
  this.addStatic(data.static);
  this.addTableAttributes(data.attributes);
  console.log('Creating ModelFactory: ', this.tableName);
  return this;
};

ModelFactory.prototype.createModel = function(methods) {

  //
  // Create base
  var Model = function() {
    var created = false;
    this.getCreated = function() {
      return created;
    };
    this.setCreated = function(isCreated) {
      created = isCreated;
    };
  };

  Model.prototype.save = function(data) {
    // TODO
  };

  Model.prototype.delete = function(data) {
    // TODO
  };

  //
  // Add methods to model prototype
  for(var method in methods) {
    Model.prototype[method] = methods[method];
  }

  //
  // Return
  return Model;
};

ModelFactory.prototype.addPre = function(pre) {
  for(method in pre) {
    this.pre[method] = pre[method];
  }
}

ModelFactory.prototype.addPost = function(post) {
  for(method in post) {
    this.post[method] = post[method];
  }
}

ModelFactory.prototype.addStatic = function(static) {
  for(method in static) {
    this[method] = static[method];
  }
}

ModelFactory.prototype.addTableAttributes = function(attributes) {
  for(var attribute in attributes) {
    switch(typeof attributes[attribute]) {
      case 'object': {
        this.attributes[attribute] = attributes[attribute];
        break;
      }
      case 'function': {
        this.relationships[attribute] = attributes[attribute](this);
        break;
      }
    }
  }
};

//
// Model Instance Operations
//

ModelFactory.prototype.initializeModel = function(data) {
  var model = new this.Model();
  // Set all attributes
  for(attribute in data) {
    model[attribute] = data[attribute];
  }
  // Set all model methods
  return model;
};

ModelFactory.prototype.initializeModels = function(data) {
  var self = this;
  var models = data.map(function(d) {
    return self.initializeModel(d);
  });
  return models;
};

//
// Attribute Operations
//

ModelFactory.prototype.getReturnAttributes = function() {
  var returnAttributes = [];
  for(var attribute in this.attributes) {
    if(!this.attributes[attribute].private) {
      returnAttributes.push(attribute);
    }
  }
  return returnAttributes;
};

//
// Relationship Operations
//

//
// Private
//

// TODO: Make private
ModelFactory.prototype.includeForMany = function(models, options, relationships) {
  var promises = [];
  models.forEach(function(model) {
    promises.push(this.includeForOne(model, options, relationships));
  }, this);
  return Promise.all(promises);
};

// TODO: Make private
ModelFactory.prototype.includeForOne = function(model, options, relationships) {
  var self = this;
  var promises = [];
  options.include.forEach(function(relationshipTableName) {
    var relationship = self.relationships[relationshipTableName];
    if(self.relationships.hasOwnProperty(relationshipTableName)) {
      var promise = relationship.include(relationshipTableName, model);
      promises.push(promise);
    }
  });
  return Promise.all(promises).then(function(values) {
    return model;
  });
};

//
// Query Operations
//

ModelFactory.prototype.hook = function(hook, data) {
  return new Promise(function(resolve, reject) {
    if(hook) {
      hook(data, resolve);
    } else {
      resolve(data);
    }
  });
}

ModelFactory.prototype.create = function(data) {
  var self = this;
  // Attach belongsToOne relationships
  var query = new QueryObject().filter(data, self.relationships);
  query.attachBelongsToOneRelationships(self.relationships);
  // Pre
  return self.hook(self.pre.create, data)
  .then(function(newData) {
    data = Object.assign({}, data, newData);
  })
  // Action
  .then(function() {
    return self.ORM.query.create(self.ORM, self.tableName, query.getAttributes(), self.getReturnAttributes());
  })
  // Attach
  .then(function(rawData) {
    var model = self.initializeModel(rawData[0]);
    model.setCreated(true);
    // Attach belongsToMany relationships
    return query.attachBelongsToManyRelationships(model, self.relationships)
  })
  // Post
  .then(function(model) {
    return self.hook(self.post.create, model)
  })
  // Return
  .then(function(model) {
    return model;
  });
};

/*
ModelFactory.prototype.update = function(where, data) {
  var self = this;
  return self.ORM.query.update(self.ORM, self.tableName, where, data, self.getReturnAttributes())
  .then(function(rawData) {
    return self.initializeModel(rawData[0]);
  });
};
*/

ModelFactory.prototype.find = function(where, options) {
  var self = this;

  // Create Query
  var query = new QueryObject().filter(where, self.relationships);

  // Make search
  return self.ORM.query.findJoin(self.ORM, self.tableName, query.getRelationships(), query.getAttributes(), self.relationships, self.getReturnAttributes())
  .then(function(rawData) {
    return self.initializeModels(rawData);
  })
  .then(function(models) {
    if(options && options.include) {
      return self.includeForMany(models, options, query.relationships);
    } else {
      return models;
    }
  });
};

ModelFactory.prototype.findOne = function(where) {
  var self = this;
  return self.ORM.query.find(self.ORM, self.tableName, where, self.getReturnAttributes())
  .then(function(rawData) {
    return (rawData.length) ? self.initializeModel(rawData[0]) : null;
  });
};

ModelFactory.prototype.findOrCreate = function(where) {
  var self = this;
  return self.findOne(where).then(function(model) {
    if(model) {
      return model;
    } else {
      return self.create(where).then(function(model) {
        return model;
      });
    }
  });
};

ModelFactory.prototype.delete = function(where) {
  var self = this;
  return self.ORM.query.delete(self.ORM, self.tableName, where)
  .then(function(rawData) {
    return rawData;
  });
};

//
// Relationship
//

ModelFactory.prototype.belongsToOne = function(modelFactoryName, parentIdentifer, childIdentifer) {
  return new Relationship({
    ORM: this.ORM,
    type: Relationship.Types.belongsToOne,
    modelFactoryName: modelFactoryName,
    parentIdentifer: parentIdentifer,
    childIdentifer: childIdentifer
  });
};

ModelFactory.prototype.belongsToMany = function(modelFactoryName, throughTableName,
  parentIdentifer_0, childIdentifer_0, parentIdentifer_1, childIdentifer_1) {
  return new Relationship({
    ORM: this.ORM,
    type: Relationship.Types.belongsToMany,
    modelFactoryName: modelFactoryName,
    parentIdentifer_0: parentIdentifer_0,
    childIdentifer_0: childIdentifer_0,
    parentIdentifer_1: parentIdentifer_1,
    childIdentifer_1: childIdentifer_1,
    throughTableName: throughTableName
  });
};

ModelFactory.prototype.hasMany = function(modelFactoryName, parentIdentifer, childIdentifer) {
  return new Relationship({
    ORM: this.ORM,
    type: Relationship.Types.hasMany,
    modelFactoryName: modelFactoryName,
    parentIdentifer: parentIdentifer,
    childIdentifer: childIdentifer
  });
};

ModelFactory.prototype.hasOne = function(modelFactoryName, parentIdentifer, childIdentifer) {
  return new Relationship({
    ORM: this.ORM,
    type: Relationship.Types.hasOne,
    modelFactoryName: modelFactoryName,
    parentIdentifer: parentIdentifer,
    childIdentifer: childIdentifer
  });
};


module.exports = ModelFactory;